(function() {
    if (window.requestIdleCallback == null) {
        // 模拟 requestIdleCallback 函数
        function mockRequestIdleCallback(callback, options = {}) {
            const timeout = (options.timeout == null || options.timeout == undefined) ? 32 : options.timeout // 默认超时时间，可根据需要调整
            const maxDeadline = Date.now() + timeout

            // 创建一个模拟的 IdleDeadline 对象
            const mockDeadline = {
                didTimeout: false,
                timeRemaining() {
                    const currentTime = Date.now()
                    if (currentTime >= maxDeadline) {
                        this.didTimeout = true
                        return 0
                    }
                    return Math.max(0, maxDeadline - currentTime)
                },
            }

            // 使用 setTimeout 模拟空闲回调
            return setTimeout(() => {
                callback(mockDeadline)
            }, 0)
        }

        // 取消逻辑可以简单地忽略，因为setTimeout已经触发
        // 或者尝试清除定时器，但可能已执行，故实际效果有限
        function mockCancelIdleCallback(id) {
            clearTimeout(id)
        }

        window.requestIdleCallback = window.requestIdleCallback || mockRequestIdleCallback
        window.cancelIdleCallback = window.cancelIdleCallback || function(id) { clearTimeout(id) }
    }
})();

(function() {
    if (window.PositionObserver == null) {
        /**
         * 比较2个位置数据是否一致
         * @param {IPosition} p1 位置1
         * @param {IPosition} p2 位置2
         * @returns {boolean}
         */
        function isSamePosition(p1, p2) {
            if (p1.bottom !== p2.bottom) return false
            if (p1.top !== p2.top) return false
            if (p1.left !== p2.left) return false
            if (p1.right !== p2.right) return false
            if (p1.width !== p2.width) return false
            if (p1.height !== p2.height) return false
            return true
        }

        class PositionObserver {
            /** @type {IntersectionObserver} */
            _intersectionObserver = null;
            /** @type {number[]} */
            _runner = [];
            /** @type {WeakMap<object, IRecord>} */
            _records = new WeakMap();
            /**
            * 更新记录
            * @param {HTMLElement} targetEl 观测DOM
            * @param {Partial<IRecord>} partialData 需要更新的数据
            */
            _updateRecord = (targetEl, partialData) => {
                const data = this._records.get(targetEl)
                const updatedData = Object.assign({}, data, partialData)
                this._records.set(targetEl, updatedData)
            }

            constructor() {
                this._intersectionObserver = new IntersectionObserver((entries) => {
                    entries.forEach((entry) => {
                        const targetEl = entry.target
                        const isIntersecting = typeof entry.isIntersecting === 'boolean' ? entry.isIntersecting : entry.intersectionRatio > 0
                        if (isIntersecting) {
                            this._startObserve(targetEl)
                        } else {
                            this._stopObserve(targetEl)
                        }
                    })
                })
            }

            /**
            * 开启元素位置实时获取（idle）
            * @param {HTMLElement} targetEl 待观测的DOM元素
            * @returns
            */
            _startObserve = (targetEl) => {
                const record = this._records.get(targetEl)
                if (!record) return
                if (record.cancelIdleId) return
                const { callback } = record
                const computeDOMPosition = () => {
                    const rect = targetEl.getBoundingClientRect()
                    const { top, right, bottom, left, width, height } = rect
                    const record = this._records.get(targetEl)
                    const prev = record ? record.position : undefined
                    const position = { top, right, bottom, left, width, height }
                    if (!prev || !isSamePosition(prev, position)) {
                        callback(targetEl, position, this)
                    }
                    this._updateRecord(targetEl, { position })
                };

                /**
                 * 在页面空闲时间计算目标元素的位置信息
                 * @type IdleRequestCallback
                 */
                const computeDOMPositionWhenIdle = (deadline) => {
                    if (deadline.didTimeout) return
                    while (deadline.timeRemaining() > 0) {
                        computeDOMPosition()
                    }
                    const cancelIdleId = requestIdleCallback(computeDOMPositionWhenIdle)
                    this._updateRecord(targetEl, { cancelIdleId })
                }

                const cancelIdleId = requestIdleCallback(computeDOMPositionWhenIdle)
                this._updateRecord(targetEl, { cancelIdleId })
                this._runner.push(cancelIdleId)
            }

            /**
            * 停止元素位置实时获取(idle)
            * @param {HTMLElement} targetEl 待观测的DOM元素
            * @returns
            */
            _stopObserve = (targetEl) => {
                const record = this._records.get(targetEl)
                if (!record) return
                const { cancelIdleId } = record
                if (!cancelIdleId) return
                cancelIdleCallback(cancelIdleId)
                this._runner = this._runner.filter(id => id !== cancelIdleId)
                this._updateRecord(targetEl, { cancelIdleId: 0 })
            }

            /**
            * 开始观察元素的位置变化
            * @param {HTMLElement|string} target 需要观察的DOM元素
            * @param {IObserveCallback} callback 对象位置变化时的回调函数
            */
            observe(target, callback) {
                /** @type HTMLElement */
                const targetEl = target instanceof HTMLElement ? target : document.querySelector(target)
                if (!targetEl) {
                    console.error(new Error('观察元素无法找到'))
                    return
                }
                if (this._records.has(targetEl)) {
                    console.error(new Error('元素已经被观察，请不要重复观测'))
                    return
                } else {
                    // 初始化位置信息
                    const rect = targetEl.getBoundingClientRect()
                    const { top, right, bottom, left, width, height } = rect
                    const position = { top, right, bottom, left, width, height }
                    callback(targetEl, position, this)
                    this._updateRecord(targetEl, { position, callback })
                }

                this._intersectionObserver.observe(target)
            }

            unobserve(target) {
                /** @type {HTMLElement} */
                const targetEl = target instanceof HTMLElement ? target : document.querySelector(target)
                if (!targetEl) return
                const record = this._records.get(targetEl)
                if (!record) return
                if (record.cancelIdleId) {
                    cancelIdleCallback(record.cancelIdleId)
                    this._runner = this._runner.filter(id => id !== record.cancelIdleId)
                }
                this._records.delete(targetEl)
                this._intersectionObserver.unobserve(targetEl)
            }

            disconnect() {
                this._runner.forEach(cancelIdleId => cancelIdleCallback(cancelIdleId))
                this._runner = []
                this._records = new WeakMap()
                this._intersectionObserver.disconnect()
            }
        }

        window.PositionObserver = PositionObserver;
    }
})();
